#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import logging
from optparse import OptionParser
from itertools import *
from subprocess import *

def l_d(smth):
    logging.debug(str(smth))
    
def command(params,stdin=None):
    l_d(params)
    process=Popen(params,stdin=PIPE, stdout=PIPE, stderr=PIPE, universal_newlines=True)
    out=process.communicate(stdin)
    if out[1]: l_d(out[1])
    if process.returncode != 0: raise Exception("*** External program failed: %s" % params[0])
    return out[0]

def flatten(listOfLists): 
    return list(chain.from_iterable(listOfLists))

def hdr_parms(header, knd, sep='/'): 
    'filter header for params starting with "knd/" '
    return [item[len(knd)+len(sep):] for item in header if item.startswith(knd+sep)]

def hdr_subparm(param, knd, default=None): 
    out=hdr_parms(param, knd, sep='=')
    return out[0] if out else default
    
def hdr_read(kap_name): 
    'read KAP header'
    f=open(kap_name, 'r')
    header=[]
    for l in takewhile( lambda x: x.find('\x1A') == -1, f):
        if l.startswith(' '):
            header[-1]=header[-1]+','+(l.strip())
        else:
            header.append(l.strip())
    del f
    if not (header and header[0].startswith('!') and hdr_parms(header,'BSB') and hdr_parms(header,'KNP')): 
        raise Exception("*** invalid file: %s" % kap_name)
    if not hdr_parms(header, 'CRR'):
        header[0]='CRR/'+header[0][header[0].find(' ')+1:]
    l_d(header)
    return header

def projection(bsb_header):
    'returns srs for the BSB chart projection and srs for the REF points'
    proj_map={
        'MERCATOR':
            ['+proj=merc +lon_0=%s',['PP'],['0']],
        'POLYCONIC':
            ['+proj=poly +lon_0=%s',['PP'],['0']],
        'TRANSVERSE MERCATOR':
            ['+proj=tmerc +lon_0=%s +lat_0=%s +k=%s +y_0=%s +x_0=%s',['PP','P1','P2','P3','P4'],['0','0','1','0','0']]
         }
    projdef=hdr_parms(bsb_header, 'KNP')[0].split(',')
    prtype=hdr_subparm(projdef, 'PR')
    if not prtype in proj_map:
        raise Exception("*** Unsupported projection: %s" % prtype)
    # get projection parameters
    parms=[ hdr_subparm(projdef, i[0], i[1]) 
        for i in zip(proj_map[prtype][1], proj_map[prtype][2]) ] 
    datum='+datum=WGS84'
    srs=(proj_map[prtype][0] % tuple(parms)) + ' ' + datum
    l_d(srs)
    return srs

def kap2vrt(kap):
    (map_dir, map_fname)=os.path.split(kap)
    base=os.path.join(options.dest_dir, os.path.splitext(map_fname)[0])
    aux_vrt= base+'.aux.vrt'    # auxilary VRT file
    out_vrt= base+'.vrt'        # output VRT file

    hdr=hdr_read(kap)           # Read BSB header
    out_srs=projection(hdr)

    l=map(eval,hdr_parms(hdr, 'DTM')) # Get northing/easting data
    dtm=l[0] if l else (0,0)
    l_d(dtm)

    refs=map(eval,hdr_parms(hdr, 'REF'))  # Get a list of geo refs in tuples
    l_d(refs)
    latlong=''.join(['%f %f\n' % (ref[4]+dtm[1]/3600,ref[3]+dtm[0]/3600) for ref in refs])
    l_d(latlong)
    refs_merc=[ i.split() for i in 
        command(['proj'] + out_srs.split(), latlong).splitlines()]
    l_d(refs_merc)
    gcps=flatten([('-gcp', `i[0][1]`,`i[0][2]`,i[1][0],i[1][1]) for i in izip(refs, refs_merc)])
    out=command(['gdal_translate','-expand','rgb','-of','VRT',kap, aux_vrt,'-a_srs',out_srs] + gcps)
    print out

    plys=map(eval,hdr_parms(hdr, 'PLY')) # Gather border polygon coordinates
    if options.cut and plys: # polygon is there, then cut
        if os.path.exists(out_vrt): os.remove(out_vrt)
        # Create cutline
        latlong=''.join(['%f %f\n' % (i[2],i[1]) for i in plys])
        poly_tr=command(['gdaltransform', '-i', '-t_srs', '+proj=longlat +datum=WGS84', kap], latlong).splitlines()
        poly='POLYGON(('+','.join(poly_tr)+'))'
        command(['gdalwarp','-wo', 'INIT_DEST=NO_DATA','-wo', 'CUTLINE='+poly,'-dstalpha','-of','VRT', aux_vrt, out_vrt])
    else:
        os.rename(aux_vrt, out_vrt)

if __name__=='__main__':
    usage = "usage: %prog [--cut] [--dest-dir=DEST_DIR] KAP_file..."
    parser = OptionParser(usage=usage,
        description="Extends builtin GDAL's support for BSB charts: WGS84 northing/easting, more projections, border line. "
        "The script converts .kap file with into GDAL .vrt, optionally clipping it out accroding to BSB border line.")
    parser.add_option("-c", "--cut", action="store_true", 
        help='cut the raster as per BSB border line')
    parser.add_option("-d", "--dest-dir", dest="dest_dir", default='',
        help='destination directory (default: current)')
    parser.add_option("-v", "--verbose", action="store_true", dest="verbose")

    (options, args) = parser.parse_args()
    if not args:
        parser.error('No input file(s) specified')
    logging.basicConfig(level=logging.DEBUG if options.verbose else logging.INFO)

    for f in args:
        kap2vrt(f)

