#!/usr/bin/env python

import os
import logging
from optparse import OptionParser
from itertools import *
from subprocess import *

#logging.basicConfig(level=logging.INFO)
logging.basicConfig(level=logging.DEBUG)

parser = OptionParser()
usage = "usage: %prog [options] arg"
(options, args) = parser.parse_args()

def l_d(smth):
    logging.debug(str(smth))
    
def command(params,stdin=None):
    process=Popen(params,stdin=PIPE, stdout=PIPE, stderr=PIPE, universal_newlines=True)
    out=process.communicate(stdin)
    if out[1]: l_d(out[1])
    if process.returncode != 0: raise Exception("*** External program failed: %s" % params[0])
    return out[0]

def flatten(listOfLists): 
    return list(chain.from_iterable(listOfLists))

def hdr_parms(header, knd, eval_item=True): 
    'filter header for params starting with "knd/" '
    return [eval(item[len(knd)+1:]) if eval_item else item[len(knd)+1:].split(',')
        for item in header if item.startswith(knd+'/')]

def hdr_read(kap_name): 
    'read KAP header'
    f=open(kap_name, 'r')
    header=[]
    for l in takewhile( lambda x: x.find('\x1A') == -1, f):
        if l.startswith(' '):
            header[-1]=header[-1]+','+(l.strip())
        else:
            header.append(l.strip())
    del f
    if not header or not header[0].startswith('!') or not hdr_parms(header, 'BSB', False): 
        raise Exception("*** invalid file: %s" % kap_name)
    if not hdr_parms(header, 'CRR', False):
        header[0]='CRR/'+header[0][header[0].find(' ')+1:]
    l_d(header)
    return header

for kap in args:
    base=os.path.splitext(kap)[0]
    aux_vrt= base+'.aux.vrt' # auxilary VRT file
    out_vrt= base+'.vrt' # output VRT file
    
    hdr=hdr_read(kap)
    # Get northing/easting data
    lines=hdr_parms(hdr, 'DTM')
    if lines:
        dtm=lines[0]
        l_d(dtm)
    else:
        dtm=(0,0)

    refs=hdr_parms(hdr, 'REF') # get a list of refs in tuples
    l_d(refs)
    # WGS84:
    # geographic: +proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs
    # Mercator:   +proj=merc +lon_0=0 +k=1 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units=m +no_defs
    wgs84_latlong='+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs'    
    wgs84_merc='+proj=merc +lon_0=0 +k=1 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units=m +no_defs'    
    latlong=''.join(['%f %f\n' % (ref[4]+dtm[1]/3600,ref[3]+dtm[0]/3600) for ref in refs])
    l_d(latlong)
    refs_merc=[ i.split() for i in 
        command(['gdaltransform', '-s_srs', wgs84_latlong, '-t_srs', wgs84_merc], latlong).splitlines()]
    l_d(refs_merc)
    gcps=flatten([ ('-gcp', `i[0][1]`,`i[0][2]`,i[1][0],i[1][1]) for i in izip(refs, refs_merc)])
    gtrans_args=[ 'gdal_translate', '-expand', 'rgb', '-of', 'VRT', kap, aux_vrt, '-a_srs', wgs84_merc ]+gcps
    l_d(gtrans_args)
    out=command(gtrans_args)
    print out

    plys=hdr_parms(hdr, 'PLY') # Gather polygon coordinates
    l_d(plys)
    if plys:
        gdal_16 = command(['gdalwarp', '--version']).startswith('GDAL 1.6.')
        # Create cutline
        cutline= base+'.cut.gmt'
        f = open(cutline, 'w')
        if gdal_16:
            print >> f, \
                '# @VGMT1.0 @GPOLYGON\n'\
                '# FEATURE_DATA\n'\
                '>\n'\
                '# @P'
            latlong=''.join(['%f %f\n' % (i[2],i[1]) for i in plys])
            poly_tr=[ i.split() for i in command(['gdaltransform', '-i', '-t_srs', 'WGS84', kap], latlong).splitlines()]
        else:
            print >> f, \
                '# @VGMT1.0 @GPOLYGON\n'\
                '# @Jp"%s"\n'\
                '# FEATURE_DATA\n'\
                '>\n'\
                '# @P' % wgs84_merc
            latlong=''.join(['%f %f\n' % (i[2]+dtm[1]/3600,i[1]+dtm[0]/3600) for i in plys])
            poly_tr=[ i.split() for i in command(['gdaltransform','-s_srs',wgs84_latlong,'-t_srs',wgs84_merc],latlong).splitlines()]
        l_d(latlong)
        for i in poly_tr:
            print >> f,"%s %s" % (i[0],i[1])
        del f
        if os.path.exists(out_vrt): os.remove(out_vrt)
        out=command(['gdalwarp','-wo', 'INIT_DEST=NO_DATA','-dstalpha','-cutline',cutline,'-of','VRT', aux_vrt, out_vrt])
        l_d(out)
    else:
        os.rename(aux_vrt, out_vrt)

